.page-header {
  background-color: #051f2d;
  text-align: center;
  font-size: 30px;
  color: #fff;
  border-radius: 0;
}

strong {
  color: #f25d47;
  font-weight: 700;
  white-space: nowrap;
}

.page-content {
  padding: 0 30px;
}
